package com.rizzoweb.alarm;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

/**
  * This class implements an alarm manager similar to Unix <code>cron</code>
  * and <code>at</code> daemons. It is intended to fire events
  * when alarms' date and time match the current ones. Alarms are
  * added dynamically and can be one-shot or repetitive
  * (i.e. rescheduled when matched). Time unit is seconds. Alarms
  * scheduled less than one second to the current time are rejected (a
  * <code>PastDateException</code> is thrown).<p>
  *
  * The alarm scheduler has been designed to
  * manage a large quantity of alarms and to reduce the use of the CPU
  * time (the AlarmManager's thread is started only when there are
  * alarms to be managed and it sleeps until the next alarm
  * date). <p>
  *
  * Note : because of clocks' skews some alarm dates may be erroneous,
  * particularly if the next alarm date is scheduled for a remote time
  * (e.g. more than a few days). In order to avoid that problem,
  * well-connected machines can use the Network Time
  * Protocol (NTP) to synchronize their clock.<p>
  *
  * Example of use:
  * <pre>
  *  // Creates a new AlarmManager
  *  AlarmManager mgr = new AlarmManager();
  *
  *  // Date alarm (non repetitive)
  *  mgr.addAlarm(new Date(System.currentTimeMillis() + 300000),
  *               new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("5 minutes later");
  *    }
  *  });
  *
  *  Calendar cal = Calendar.getInstance();
  *  cal.add(Calendar.WEEK_OF_YEAR, 1);
  *  mgr.addAlarm(cal.getTime(), new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("One week later");
  *    }
  *  });
  *
  *  // Alarm with a delay (in minute) relative to the current time.
  *  mgr.addAlarm(1, true, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("1 more minute ! (" + new Date() + ")");
  *    }
  *  });
  *
  *  // Cron-like alarm (minute, hour, day of month, month, day of week, year)
  *  // Repetitive when the year is not specified.
  *
  *  mgr.addAlarm(-1, -1, -1, -1, -1, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("Every minute (" + new Date() + ")");
  *    }
  *  });
  *
  *  mgr.addAlarm(5, -1, -1, -1, -1, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("Every hour at 5' (" + new Date() + ")");
  *    }
  *  });
  *
  *  mgr.addAlarm(00, 12, -1, -1, -1, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("Lunch time (" + new Date() + ")");
  *    }
  *  });
  *
  *  mgr.addAlarm(07, 14, 1, Calendar.JANUARY, -1, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("Happy birthday Lucas !");
  *    }
  *  });
  *
  *  mgr.addAlarm(30, 9, 1, -1, -1, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("On the first of every month at 9:30");
  *    }
  *  });
  *
  *  mgr.addAlarm(00, 18, -1, -1, Calendar.FRIDAY, -1, new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("On every Friday at 18:00");
  *    }
  *  });
  *
  *  mgr.addAlarm(00, 13, 1, Calendar.AUGUST, -1, 2001,  new AlarmListener() {
  *    public void handleAlarm(AlarmSchedule schedule) {
  *      System.out.println("2 years that this class was programmed !");
  *    }
  *  });
  * </pre>
  */

public class AlarmManager {

	protected AlarmWaiter waiter;
	protected SortedSet queue;
	private boolean debug = true;

	private void debug(String s) {
	  if (debug)
		System.out.println("[" + Thread.currentThread().getName() + "] AlarmManager: " + s);
	}

	/**
	  * Creates a new AlarmManager. The waiter thread will be started
	  * only when the first alarm listener will be added.
	  *
	  * @param isDaemon true if the waiter thread should run as a daemon.
	  * @param threadName the name of the waiter thread
	  */
	public AlarmManager(boolean isDaemon, String threadName) {
	  queue = new TreeSet();
	  waiter = new AlarmWaiter(this, isDaemon, threadName);
	 }

	/**
	 * Creates a new AlarmManager. The waiter thread will be started
	 * only when the first alarm listener will be added. The waiter
	 * thread will <i>not</i> run as a daemon.
	  */
	public AlarmManager() {
	  this(false, "AlarmManager");
	}

	/**
	 * Returns a copy of all alarms (instances of AlarmSchedule) in the manager.
	 */
	public synchronized List getAllAlarms() {

		LinkedList result = new LinkedList();
		Iterator iterator = queue.iterator();

		while (iterator.hasNext()) {
			result.add(iterator.next());
		}

		return result;
	}

	/**
	 * Notifies the appropriate AlarmListener(s) when an alarm date is
	 * reached. This method is only to be called by the the AlarmWaiter or
	 * myself (if the next alarm is in less than 1 second.)
	 */
	protected synchronized void notifyListeners() {

		debug("Alarm notification");

		if (queue.isEmpty()) {
			return;
		}

		AlarmSchedule schedule = (AlarmSchedule) queue.first();
		queue.remove(schedule);
		schedule.notifyListener();

		if (schedule.isRepetitive()) {
			schedule.updateAlarmTime();
			queue.add(schedule);
		}

		if (queue.isEmpty()) {
			debug("Queue is empty");
		} else {
			long alarmTime = ((AlarmSchedule) queue.first()).getAlarmTime();
			if (alarmTime - System.currentTimeMillis() < 1000) {
				debug("The next alarm is very close - notifying now");
				notifyListeners();
			} else {
				debug("Updating the waiter for " + queue.first());
				waiter.restart(alarmTime);
			}
		}
	}

	/**
	  * Adds an alarm for the specified date and time.
	  *
	  * @param minute minute of the alarm (0-59).
	  * @param hour hour of the alarm (0-23).
	  * @param dayOfMonth day of month of the alarm (1-31, -1 for every day)
	  *		This attribute is exclusive with dayOfWeek.
	  * @param month month of the alarm (0-11, -1 for every month).
	  *		java.util.Calendar constants can be used.
	  * @param dayOfWeek day of week of the alarm (1-7, -1 if every day,
	  *		1=Sunday, 7=Saturday). java.util.Calendar constants can be
	  *		used.  This attribute is exclusive with dayOfMonth.
	  * @param year year of the alarm. When this field is set to -1,
	  *		the alarm is repetitive.
	  * @param listener the listener to be notified when the alarm fires.
	  * @exception PastDateException if the alarm date is in the past
	  *		or less than 1 second after the current time.
	  */
	public synchronized void addAlarm(int minute, int hour,
		int dayOfMonth, int month, int dayOfWeek, int year,
		AlarmListener listener) throws PastDateException {

		addAlarm(new AlarmSchedule(minute, hour, dayOfMonth, month,
								dayOfWeek, year, listener));
	}

	/**
	 * Adds an alarm for a specified delay.
	 *
	 * @param delay The delay time in minutes (relative to now).
	 * @param isRepetitive <code>true</code> if the alarm should repeat.
	 * @param listener The listener to be notified when the alarm fires.
	 * @exception PastDateException if the alarm date is in the past
	 *		or less than 1 second after the current time.
	 */
	public synchronized void addAlarm(int delay, boolean isRepetitive,
		AlarmListener listener) throws PastDateException {

		addAlarm(new AlarmSchedule(delay, isRepetitive, listener));
	}

	/**
	 * Adds an alarm for the specified date.  The given listener will be
	 * notified when the alarm fires.
	 */
	public synchronized void addAlarm(Date date,
		AlarmListener listener) throws PastDateException {

		AlarmSchedule schedule = new AlarmSchedule(date, listener);
		addAlarm(schedule);
	}

	/**
	 * Adds an alarm for the specified AlarmSchedule
	 */
	public synchronized void addAlarm(AlarmSchedule schedule) {

		debug("Adding alarm: " + schedule);

		if (queue.add(schedule)) {
			debug("This new alarm is the top one, update the waiter thread");
			waiter.update(((AlarmSchedule) queue.first()).getAlarmTime());
		}

		debug("New queue size is " + queue.size());
	}

	/**
	 * Answers whether or not I contain the given AlarmSchedule.
	 */
	public synchronized boolean containsAlarm(AlarmSchedule schedule) {
	  return queue.contains(schedule);
	}

	/**
	 * Removes the specified AlarmSchedule.
	 */
	public synchronized void removeAlarm(AlarmSchedule schedule) {

		debug("Removing alarm: " + schedule);

		if (queue.remove(schedule)) {
			if (!queue.isEmpty()) {
				waiter.update(((AlarmSchedule) queue.first()).getAlarmTime());
			}
		}

		debug("New queue size is " + queue.size());
	}
}
