package com.rizzoweb.alarm;

import java.util.Calendar;

/**
 * This is an Alarmschedule that repeats itself every day at a specified
 * time.  Clients of this class can specify days that are excluded
 * (ie., days when an alarm will not be triggered).
 *
 * @version 1.2
 */
public class DailyAlarmSchedule extends AlarmSchedule {

	public static final String version = "1.2";

	/**
	 * Boolean array indicating which days are excluded from firing
	 * an alarm.  If an index has a value of true, then no alarm will
	 * be fired on the corresponding day (1=Sunday, ..., 7=Saturday).
	 * @see excludeDay()
	 * @see setDayExcluded()
	 */
	private boolean[] excludedDays;

	/**
	 * Updates my alarm time so that I am rescheduled.
	 */
	protected void updateAlarmTime() {

		Calendar now = Calendar.getInstance();
		Calendar alarm = (Calendar) now.clone();
		int nowHour = now.get(Calendar.HOUR_OF_DAY);
		int nowMinute = now.get(Calendar.MINUTE);

		alarm.set(Calendar.HOUR_OF_DAY, getHour());
		alarm.set(Calendar.MINUTE, getMinute());
		alarm.set(Calendar.SECOND, 0);


		// Increment dayOfYear if "now" is past or equal to the alarm:
		if (nowHour > getHour() ||
			(nowHour == getHour() && (nowMinute >= getMinute()))) {
				alarm.roll(Calendar.DAY_OF_YEAR, true);
		}

		// Skip excluded days:
		while (isDayExcluded(alarm.get(Calendar.DAY_OF_WEEK))) {
			alarm.roll(Calendar.DAY_OF_YEAR, true);
		}

		setAlarmTime(alarm.getTime().getTime());
	}

	/**
	 * Constructor is private to force usage of the factory methods.
	 */
	private DailyAlarmSchedule(int hour, int minute, AlarmListener aListener)
		throws PastDateException {

		super(minute, hour, -1, -1, -1, -1, aListener);
	}

	/**
	 * Adds the given day to my excludedDays list (excluded days
	 * are days on which I will not trigger an alarm).
	 * Days are numbered from 1 to 7, with 1=Sunday and 7=Saturday.
	 * java.util.Calendar constants, such as Calendar.TUESDAY, can
	 * be used as parameter values.
	 */
	public void excludeDay(int aDay) {
		setDayExcluded(aDay, true);
	}

	/**
	 * Answers my array of booleans indicating which days are excluded
	 * from triggering an alarm.
	 */
	protected boolean[] getExcludedDays() {

		if (excludedDays == null) {
			excludedDays = new boolean[]
				{false, false, false, false, false, false, false, false};
			// There are 8 entries because the Calendar day constants
			// are indexed starting at one, but arrays are zero-base
			// indexed.  Thus, the first element (index 0) is ignored.
			// I choose to do this rather than a lot of "- 1" throughout
			// the code.
		}

		return excludedDays;
	}

	/**
	 * Removes the given day from my excludedDays list (excluded days
	 * are days on which I will not trigger an alarm).
	 * Days are numbered from 1 to 7, with 1=Sunday and 7=Saturday.
	 * java.util.Calendar constants, such as Calendar.TUESDAY, can
	 * be used as parameters values.
	 */
	public void includeDay(int aDay) {
		setDayExcluded(aDay, false);
	}

	/**
	 * Answers if the given day is in my excludedDays list (excluded days
	 * are days on which I will not trigger an alarm).
	 * Days are numbered from 1 to 7, with 1=Sunday and 7=Saturday.
	 * java.util.Calendar constants, such as Calendar.TUESDAY, can
	 * be used as parameters values.
	 *
	 * @exception IndexOutOfBoundsException thrown if the specified day
	 *		number is out of range (1-7)
	 */
	public boolean isDayExcluded(int aDay) {

		verifyDayNumber(aDay);
		return getExcludedDays()[aDay];
	}

	/**
	 * Sets the exclusion flag for the given day.  On days with a true
	 * exclusion flag, I will not trigger and alarm.
	 * Days are numbered from 1 to 7, with 1=Sunday and 7=Saturday.
	 * java.util.Calendar constants, such as Calendar.TUESDAY, can
	 * be used as parameters values.
	 *
	 * @exception IndexOutOfBoundsException thrown if the specified day
	 *		number is out of range (1-7)
	 */
	public void setDayExcluded(int aDay, boolean exclude) {

		verifyDayNumber(aDay);
		getExcludedDays()[aDay] = exclude;
		updateAlarmTime();
	}

	/**
	 * Verifies that the given day number is in the valid range, 1-7.
	 * @exception IndexOutOfBoundsException thrown if the given day is
	 *		out of range.
	 */
	protected void verifyDayNumber(int aDay) {

		if (aDay < 1 || aDay > 7) {
			throw new IndexOutOfBoundsException("Day must be in the range 1-7, not " + aDay);
		}
	}

	/**
	 * Answers a new instance that repeats itself every day at the
	 * specified time.  An AlarmListener must be spcified with setListener()
	 * before the alarm is added to the manager to ensure proper handling
	 * when the alarm fires.
	 *
	 * @param minute The minute of the hour that I will trigger an alarm (0-59).
	 * @param hour The hour of the day that I will trigger an alarm (0-23).
	 */
	public static DailyAlarmSchedule newInstance(int hour, int minute) {

		// Implementation note:
		// The reason factory methods are used for this class is because
		// my superclass constructor throws an exception and but I don't
		// need to throw it.  The compiler won't let me wrap the superclass
		// constructor call in a try...catch block, so I am forced to use
		// factory methods instead.

		return newInstance(hour, minute, null);
	}

	/**
	 * Answers a new instance that repeats itself every day at the
	 * specified time.
	 *
	 * @param minute The minute of the hour that I will trigger an alarm (0-59).
	 * @param hour The hour of the day that I will trigger an alarm (0-23).
	 */
	public static DailyAlarmSchedule newInstance(int hour, int minute, AlarmListener aListener) {

		// Implementation note:
		// The reason factory methods are used for this class is because
		// my superclass constructor throws an exception and but I don't
		// need to throw it.  The compiler won't let me wrap the superclass
		// constructor call in a try...catch block, so I am forced to use
		// factory methods instead.

		DailyAlarmSchedule newSchedule;

		try {
			newSchedule = new DailyAlarmSchedule(hour, minute, aListener);
		} catch (PastDateException ex) {
			// This should never happen, since I only create recurring alarms,
			// but we have to make the compiler happy.
			newSchedule = null;
		}

		return newSchedule;
	}
}
